@@ -18,38 +18,35 @@ class N2Layout {
      * @param {boolean} showLinearSolverNames Whether to show linear or non-linear solver names.
      * @param {Object} dims The initial sizes for multiple tree elements.
      */
-    constructor(model, newZoomedElement,
-        showLinearSolverNames, dims) {
+    constructor(model, newZoomedElement, showLinearSolverNames, dims) {
         this.model = model;
 
         this.zoomedElement = newZoomedElement;
         this.showLinearSolverNames = showLinearSolverNames;
 
-        this.outputNamingType = "Absolute";
         this.zoomedNodes = [];
         this.visibleNodes = [];
 
         this.zoomedSolverNodes = [];
         this.visibleSolverNodes = [];
+        this.curVisibleNodeCount = 0;
+        this.prevVisibleNodeCount = 0;
 
         // Initial size values derived from read-only defaults
         this.size = dims.size;
         this.svg = d3.select("#svgId");
 
+        startTimer('N2Layout._computeLeaves');
+        this._computeLeaves();
+        stopTimer('N2Layout._computeLeaves');
+
         this._setupTextRenderer();
         startTimer('N2Layout._updateTextWidths');
         this._updateTextWidths();
         stopTimer('N2Layout._updateTextWidths');
 
-        startTimer('N2Layout._updateSolverTextWidths');
-        this._updateSolverTextWidths();
-        stopTimer('N2Layout._updateSolverTextWidths');
         delete (this.textRenderer);
 
-        startTimer('N2Layout._computeLeaves');
-        this._computeLeaves();
-        stopTimer('N2Layout._computeLeaves');
-
         startTimer('N2Layout._computeColumnWidths');
         this._computeColumnWidths();
         stopTimer('N2Layout._computeColumnWidths');
@@ -86,8 +83,12 @@ class N2Layout {
      * the original transition methods are restored.
      */
     setTransitionPermission() {
+        this.prevVisibleNodeCount = this.visibleNodeCount;
+        this.visibleNodeCount = this.visibleNodes.length;
+        const highWaterMark = Math.max(this.prevVisibleNodeCount, this.visibleNodeCount);
+
         // Too many nodes, disable transitions.
-        if (this.visibleNodes.length >= N2TransitionDefaults.maxNodes) {
+        if (highWaterMark >= N2TransitionDefaults.maxNodes) {
             debugInfo("Denying transitions: ", this.visibleNodes.length,
                 " visible nodes, max allowed: ", N2TransitionDefaults.maxNodes)
 
@@ -160,11 +161,11 @@ class N2Layout {
 
         let retVal = node.name;
 
-        if (this.outputNamingType == "Promoted" &&
-            node.isParamOrUnknown() &&
-            this.zoomedElement.propExists('promotions') &&
-            this.zoomedElement.promotions.propExists(node.absPathName)) {
-            retVal = this.zoomedElement.promotions[node.absPathName];
+        if (node.name == '_auto_ivc') {
+            retVal = 'Auto-IVC';
+        }
+        else if (node.absPathName.match(/^_auto_ivc.*/) && node.promotedName !== undefined) {
+            retVal = node.promotedName;
         }
 
         return retVal;
@@ -198,7 +199,12 @@ class N2Layout {
         node.nameWidthPx = this._getTextWidth(this.getText(node)) + 2 *
             this.size.rightTextMargin;
 
-        if (node.hasChildren()) {
+        if (!node.isInputOrOutput()) {
+            node.nameSolverWidthPx = this._getTextWidth(this.getSolverText(node)) + 2 *
+                this.size.rightTextMargin;
+        }
+
+        if (node.hasChildren() && !node.isMinimized) {
             for (let child of node.children) {
                 this._updateTextWidths(child);
             }
@@ -206,45 +212,35 @@ class N2Layout {
     }
 
     /**
-     * Determine text width for this and all decendents if node is a solver.
-     * @param {N2TreeNode} [node = this.zoomedElement] Item to begin looking from.
-     */
-    _updateSolverTextWidths(node = this.zoomedElement) {
-        if (node.isParam() || node.varIsHidden) {
-            return;
-        }
-
-        node.nameSolverWidthPx = this._getTextWidth(this.getSolverText(node)) + 2 *
-            this.size.rightTextMargin;
-
-        if (node.hasChildren()) {
-            for (let child of node.children) {
-                this._updateSolverTextWidths(child);
-            }
-        }
-    }
-
-    /** Recurse through the tree and add up the number of leaves that each
-     * node has, based on their array of children.
+     * Recurse through the tree and add up the number of leaves that each
+     * node has, based on their array of visible children.
      * @param {N2TreeNode} [node = this.model.root] The starting node.
      */
     _computeLeaves(node = this.model.root) {
-        if (node.varIsHidden) {
-            node.numLeaves = 0;
-        }
-        else if (node.hasChildren() && !node.isMinimized) {
-            node.numLeaves = 0;
-            for (let child of node.children) {
-                this._computeLeaves(child);
-                node.numLeaves += child.numLeaves;
+        node.numLeaves = 0;
+
+        if (!node.varIsHidden) {
+            if (node.name == '_auto_ivc' && !node.manuallyExpanded) {
+                node.minimize();
+            }
+            else if (this.model.nodeIds.length > Precollapse.minimumNodes) {
+                node.minimizeIfLarge(this.model.depthCount[node.depth]);
+            }
+
+            if (node.hasChildren() && !node.isMinimized) {
+                for (let child of node.children) {
+                    this._computeLeaves(child);
+                    node.numLeaves += child.numLeaves;
+                }
+            }
+            else {
+                node.numLeaves = 1;
             }
-        }
-        else {
-            node.numLeaves = 1;
         }
     }
 
-    /** For visible nodes with children, choose a column width
+    /**
+     * For visible nodes with children, choose a column width
      * large enough to accomodate the widest label in their column.
      * @param {N2TreeNode} node The item to operate on.
      * @param {string} childrenProp Either 'children' or 'subsystem_children'.
@@ -275,7 +271,8 @@ class N2Layout {
         }
     }
 
-    /** Compute column widths across the model, then adjust ends as needed.
+    /**
+     * Compute column widths across the model, then adjust ends as needed.
      * @param {N2TreeNode} [node = this.zoomedElement] Item to operate on.
      */
     _computeColumnWidths(node = this.zoomedElement) {
@@ -304,7 +301,8 @@ class N2Layout {
         this.cols[this.greatestDepth].width = lastColumnWidth;
     }
 
-    /** Compute solver column widths across the model, then adjust ends as needed.
+    /**
+     * Compute solver column widths across the model, then adjust ends as needed.
      * @param {N2TreeNode} [node = this.zoomedElement] Item to operate on.
      */
     _computeSolverColumnWidths(node = this.zoomedElement) {
@@ -333,9 +331,7 @@ class N2Layout {
         this.solverCols[this.greatestDepth].width = lastColumnWidth;
     }
 
-    /** Set the location of the columns based on the width of the columns
-     * to the left.
-     */
+    /** Set the location of the columns based on the width of the columns to the left. */
     _setColumnLocations() {
         this.size.partitionTree.width = 0;
         this.size.solverTree.width = 0;
@@ -388,7 +384,7 @@ class N2Layout {
             (this.cols[workNode.depth].width / this.size.partitionTree.width) : 1 - workNode.dims.x;
         node.dims.height = workNode.numLeaves / this.model.root.numLeaves;
 
-        if (node.varIsHidden) { // param or hidden leaf leaving
+        if (node.varIsHidden) { // input or hidden leaf leaving
             node.dims.x = this.cols[node.parentComponent.depth + 1].location / this.size.partitionTree.width;
             node.dims.y = node.parentComponent.dims.y;
             node.dims.width = node.dims.height = 1e-6;
@@ -425,7 +421,7 @@ class N2Layout {
                 this.zoomedSolverNodes.push(node);
             }
             if (!node.hasChildren() || node.isMinimized) { //at a "leaf" workNode
-                if (!node.isParam() && !node.varIsHidden) {
+                if (!node.isInput() && !node.varIsHidden) {
                     this.visibleSolverNodes.push(node);
                 }
                 earliestMinimizedParent = node;
@@ -443,7 +439,7 @@ class N2Layout {
 
         node.solverDims.height = workNode.numLeaves / this.model.root.numLeaves;
 
-        if (node.varIsHidden) { //param or hidden leaf leaving
+        if (node.varIsHidden) { //input or hidden leaf leaving
             node.solverDims.x = this.cols[node.parentComponent.depth + 1].location /
                 this.size.partitionTree.width;
             node.solverDims.y = node.parentComponent.dims.y;
@@ -509,9 +505,12 @@ class N2Layout {
      * @param {number} transitionStartDelay ms to wait before performing transition
      */
     updateTransitionInfo(dom, transitionStartDelay, manuallyResized) {
+
         sharedTransition = d3.transition()
             .duration(N2TransitionDefaults.duration)
-            .delay(transitionStartDelay);
+            .delay(transitionStartDelay)
+            // Hide the transition waiting animation when it ends:
+            .on('end', function () { dom.waiter.attr('class', 'no-show'); });
 
         this.transitionStartDelay = N2TransitionDefaults.startDelay;
 
@@ -599,4 +598,4 @@ class N2Layout {
 
         return { 'width': width, 'height': height };
     }
-}
\ No newline at end of file
+}
